MODULE DfConfig;

  IMPORT Strings;

  CONST
    MaxName* = 32;

    ok* = 0; errLongName* = 1; errSyn* = 2; errEos* = 3;

    CR = 0DX;

  TYPE
    Name* = ARRAY MaxName OF CHAR;
    Value* = POINTER TO ARRAY OF CHAR;

    Key* = POINTER TO LIMITED RECORD
      name*: Name;
      value-: Value;
      next-: Key
    END;

    Section* = POINTER TO LIMITED RECORD
      name*: Name;
      keys-: Key;
      next-: Section
    END;

    Config* = POINTER TO LIMITED RECORD
      sections-: Section;
      res*, pos*: INTEGER
    END;

  (* --- Key --- *)

  PROCEDURE (k: Key) GetString* (): Value, NEW;
  BEGIN
    RETURN k.value
  END GetString;

  PROCEDURE (k: Key) SetString* (IN val: ARRAY OF CHAR), NEW;
  BEGIN
    NEW(k.value, LEN(val$) + 1); k.value^ := val$;
  END SetString;

  PROCEDURE (k: Key) SetStringPtr* (val: Value), NEW;
  BEGIN
    ASSERT(val # NIL, 20);
    k.value := val
  END SetStringPtr;

  PROCEDURE (k: Key) SetInt* (x: INTEGER), NEW;
    VAR s: ARRAY 32 OF CHAR;
  BEGIN
    Strings.IntToString(x, s);
    k.SetString(s)
  END SetInt;

  PROCEDURE (k: Key) GetInt* (min, max, def: INTEGER): INTEGER, NEW;
    VAR x, res: INTEGER;
  BEGIN
    x := def;
    IF k.value # NIL THEN
      (* !!! check digits *)
      Strings.StringToInt(k.value, x, res);
      IF res = 0 THEN x := MIN(MAX(x, min), max) END
    END;
    RETURN x
  END GetInt;

  (* --- Section --- *)

  PROCEDURE (s: Section) FindKey* (IN name: Name): Key, NEW;
    VAR x: Key;
  BEGIN
    x := s.keys;
    WHILE (x # NIL) & (x.name$ # name$) DO x := x.next END;
    RETURN x
  END FindKey;

  PROCEDURE NewKey (s: Section; IN key: Name): Key;
    VAR k: Key;
  BEGIN
    k := s.keys;
    IF k = NIL THEN NEW(k); s.keys := k
    ELSIF k.name$ # key$ THEN
      WHILE (k.next # NIL) & (k.next.name$ # key$) DO k := k.next END;
      IF k.next = NIL THEN NEW(k.next) END;
      k := k.next
    END;
    k.name := key$;
    RETURN k
  END NewKey;

  PROCEDURE (s: Section) GetString* (IN key: Name): Value, NEW;
    VAR k: Key; v: Value;
  BEGIN v := NIL;
    k := s.FindKey(key);
    IF k # NIL THEN v := k.GetString() END;
    RETURN v
  END GetString;

  PROCEDURE (s: Section) SetStringPtr* (IN key: Name; v: Value), NEW;
    VAR k: Key;
  BEGIN
    k := NewKey(s, key);
    k.SetStringPtr(v)
  END SetStringPtr;

  PROCEDURE (s: Section) SetString* (IN key: Name; x: ARRAY OF CHAR), NEW;
    VAR k: Key;
  BEGIN
    k := NewKey(s, key);
    k.SetString(x)
  END SetString;

  PROCEDURE (s: Section) SetInt* (IN key: Name; x: INTEGER), NEW;
    VAR k: Key;
  BEGIN
    k := NewKey(s, key);
    k.SetInt(x)
  END SetInt;

  PROCEDURE (s: Section) GetInt* (IN key: Name; min, max, def: INTEGER): INTEGER, NEW;
    VAR k: Key; x: INTEGER;
  BEGIN x := def;
    k := s.FindKey(key);
    IF k # NIL THEN x := k.GetInt(min, max, def) END;
    RETURN x
  END GetInt;

  (* --- Config --- *)

  PROCEDURE NewSection (c: Config; IN section: Name): Section;
    VAR s: Section;
  BEGIN
    s := c.sections;
    IF s = NIL THEN NEW(s); c.sections := s
    ELSIF s.name$ # section$ THEN
      WHILE (s.next # NIL) & (s.next.name$ # section$) DO s := s.next END;
      IF s.next = NIL THEN NEW(s.next) END;
      s := s.next
    END;
    s.name := section$;
    RETURN s
  END NewSection;

  PROCEDURE (c: Config) FindSection* (IN name: Name): Section, NEW;
    VAR x: Section;
  BEGIN
    x := c.sections;
    WHILE (x # NIL) & (x.name$ # name$) DO x := x.next END;
    RETURN x
  END FindSection;

  PROCEDURE (c: Config) GetString* (IN section, key: Name): Value, NEW;
    VAR s: Section; k: Key; v: Value;
  BEGIN v := NIL;
    s := c.FindSection(section);
    IF s # NIL THEN v := s.GetString(key) END;
    RETURN v
  END GetString;

  PROCEDURE (c: Config) SetStringPtr* (IN section, key: Name; v: Value), NEW;
  BEGIN
    NewSection(c, section).SetStringPtr(key, v)
  END SetStringPtr;

  PROCEDURE (c: Config) SetString* (IN section, key: Name; x: ARRAY OF CHAR), NEW;
  BEGIN
    NewSection(c, section).SetString(key, x)
  END SetString;

  PROCEDURE (c: Config) GetInt* (IN section, key: Name; min, max, def: INTEGER): INTEGER, NEW;
    VAR s: Section; x: INTEGER;
  BEGIN x := def;
    s := c.FindSection(section);
    IF s # NIL THEN x := s.GetInt(key, min, max, def) END;
    RETURN x
  END GetInt;

  PROCEDURE (c: Config) SetInt* (IN section, key: Name; x: INTEGER), NEW;
  BEGIN
    NewSection(c, section).SetInt(key, x)
  END SetInt;

  PROCEDURE Parse* (IN src: ARRAY OF CHAR): Config;
    VAR pos, err: INTEGER; c: Config; section, key: Name; val: ARRAY 1024 OF CHAR;

    PROCEDURE Error (code: INTEGER);
    BEGIN
      err := code;
      IF c.res = ok THEN c.res := code; c.pos := pos END
    END Error;

    PROCEDURE SkipSpace;
    BEGIN
      WHILE (src[pos] # 0X) & (src[pos] <= " ") DO INC(pos) END
    END SkipSpace;

    PROCEDURE ReadName (VAR n: ARRAY OF CHAR; term: CHAR);
      VAR i: INTEGER; done: BOOLEAN;
    BEGIN
      i := 0; done := FALSE;
      WHILE ~done & (src[pos] # 0X) DO
        IF term < " " THEN done := src[pos] <= term
        ELSE done := src[pos] = term
        END;
        IF ~done THEN
          IF i < LEN(n) - 1 THEN n[i] := src[pos]; INC(i); n[i] := 0X
          ELSE Error(errLongName)
          END
        END;
        INC(pos)
      END;
      IF ~done THEN Error(errEos) END;
      SkipSpace;
    END ReadName;

    PROCEDURE Sync;
    BEGIN
      WHILE (src[pos] # 0X) & (src[pos] # "[") DO INC(pos) END
    END Sync;

  BEGIN
    NEW(c); section := ""; key := "";
    pos := 0; err := ok; c.res := ok;
    SkipSpace;
    WHILE src[pos] # 0X DO
      SkipSpace;
      IF src[pos] = "[" THEN
        INC(pos); ReadName(section, "]")
      ELSIF src[pos] = ";" THEN
        WHILE (src[pos] # 0X) & (src[pos] > CR) DO INC(pos) END
      ELSIF Strings.IsAlphaNumeric(src[pos]) THEN
        ReadName(key, "=");
        IF err = ok THEN
          ReadName(val, CR);
          IF err = ok THEN c.SetString(section, key, val) END;
        ELSE Sync
        END
      ELSE Error(errSyn); Sync
      END
    END;
    RETURN c
  END Parse;

  PROCEDURE New* (): Config;
    VAR c: Config;
  BEGIN
    NEW(c);
    RETURN c
  END New;

END DfConfig.
